@@ -13,7 +13,7 @@
  *
  */
 
-/* RCSID("$Id: ssh.h,v 1.46 2000/05/17 08:20:15 markus Exp $"); */
+/* RCSID("$OpenBSD: ssh.h,v 1.49 2000/08/19 18:48:11 markus Exp $"); */
 
 #ifndef SSH_H
 #define SSH_H
@@ -450,6 +450,9 @@ char   *tilde_expand_filename(const char *filename, uid_t my_uid);
 /* remove newline at end of string */
 char	*chop(char *s);
 
+/* return next token in configuration line */
+char	*strdelim(char **s);
+
 /* set filedescriptor to non-blocking */
 void	set_nonblock(int fd);
 
@@ -464,7 +467,7 @@ void    server_loop(pid_t pid, int fdin, int fdout, int fderr);
 void    server_loop2(void);
 
 /* Client side main loop for the interactive session. */
-int     client_loop(int have_pty, int escape_char);
+int     client_loop(int have_pty, int escape_char, int id);
 
 /* Linked list of custom environment strings (see auth-rsa.c). */
 struct envstring {
